﻿using System;
using System.Collections;
using System.Collections.Generic;

namespace HeuristicLab.Algorithms.PushGP.Stack
{
    /// <summary>
    /// While Push's stacks are generally treated as genuine stacks---that is, instructions take their arguments from the tops of
    /// the stacks and push their results onto the tops of the stacks---a few instructions (like YANK and SHOVE) do allow direct access
    /// to "deep" stack elements by means of integer indices. To this extent the stacks can be used as general, random access memory
    /// structures. This is one of the features that ensures the Turing-completeness of Push (another being the arbitrary name/value
    /// bindings supported by the NAME data type and DEFINE methods; see below).
    /// </summary>
    /// <typeparam name="T">The item type of the collection.</typeparam>
    public class PushGPStack<T> : IStack<T>
    {
        private List<T> data = new List<T>();

        public T Top => this.data[Count - 1];
        public int Count => data.Count;
        public bool IsEmpty => this.Count == 0;

        public bool IsReadOnly
        {
            get
            {
                throw new NotImplementedException();
            }
        }

        public void Add(T item)
        {
            this.Push(item);
        }

        public void Clear()
        {
            this.data.Clear();
        }

        public bool Contains(T item)
        {
            return this.data.Contains(item);
        }

        public void CopyTo(T[] array, int arrayIndex)
        {
            this.data.CopyTo(array, arrayIndex);
        }

        public IEnumerator<T> GetEnumerator()
        {
            return this.data.GetEnumerator();
        }

        public T ElementAt(int index)
        {
            return this.data[index];
        }

        public T Pop()
        {
            var index = Count - 1;
            var value = this.data[index];
            this.data.RemoveAt(index);

            return value;
        }

        public T[] Pop(int count)
        {
            var startIndex = Count - count;
            var items = this.data.GetRange(startIndex, count);

            this.data.RemoveRange(startIndex, count);

            items.Reverse();
            return items.ToArray();
        }

        public void Push(T item)
        {
            this.data.Add(item);
        }

        public void Push(params T[] items)
        {
            this.data.AddRange(items);
        }

        public void Push(IEnumerable<T> items)
        {
            this.data.AddRange(items);
        }

        public void Insert(int index, T item)
        {
            this.data.Insert(index, item);
        }

        public void Insert(int index, params T[] items)
        {
            this.data.InsertRange(index, items);
        }

        public void Insert(int index, IEnumerable<T> items)
        {
            this.data.InsertRange(index, items);
        }

        public bool Remove(T item)
        {
            var index = Count - 1;
            if (this.data.Count > 0 && this.data[index].Equals(item))
            {
                this.data.RemoveAt(index);
                return true;
            }
            else
            {
                return false;
            }
        }

        public bool TryPop(out T item)
        {
            if (this.data.Count > 0)
            {
                item = this.Pop();
                return true;
            }
            else
            {
                item = default(T);
                return false;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            return (this.data as IEnumerable).GetEnumerator();
        }

        public override string ToString()
        {
            return string.Join(", ", this.data);
        }
    }
}
